require 'ams_Lib'

module AMS
  class << self

    # Validate object type.
    # @param [Object] object
    # @param [Object, Array<Object>] types An object or an array of objects to
    #   check against.
    # @return [void]
    # @raise [TypeError] if object class doesn't match with any of the specified
    #   types.
    # @since 2.0.0
    def validate_type(object, *types)
      types = types.flatten
      return if types.empty?
      types.each { |type| return if object.is_a?(type) }
      string = case types.size
      when 1
        types[0]
      when 2
        "#{types[0]} or #{types[1]}"
      else
        "#{types[0...-1].join(', ')}, or #{types[-1]}"
      end
      raise TypeError, "Expected #{string}, but got #{object.class}.", caller
    end

    # Clamp value between min and max.
    # @param [Numeric] value
    # @param [Numeric, nil] min Pass +nil+ to have no min limit.
    # @param [Numeric, nil] max Pass +nil+ to have no max limit.
    # @return [Numeric]
    # @since 2.0.0
    def clamp(value, min, max)
      value = min if min and value < min
      value = max if max and value > max
      value
    end

    # Get numeric value sign.
    # @param [Numeric] value
    # @return [Fixnum] -1, 0, or 1
    # @since 2.0.0
    def sign(value)
      value.zero? ? 0 : (value > 0 ? 1 : -1)
    end

    # Get least value of the two values.
    # @param [Numeric] a
    # @param [Numeric] b
    # @return [Numeric]
    # @since 2.0.0
    def min(a, b)
      a < b ? a : b
    end

    # Get greatest value of the two values.
    # @param [Numeric] a
    # @param [Numeric] b
    # @return [Numeric]
    # @since 2.0.0
    def max(a, b)
      a > b ? a : b
    end

    # Scale vector.
    # @param [Array<Numeric>, Geom::Vector3d] vector
    # @param [Numeric] scale
    # @return [Geom::Vector3d]
    # @since 2.0.0
    def scale_vector(vector, scale)
      Geom::Vector3d.new(vector[0]*scale, vector[1]*scale, vector[2]*scale)
    end

    # Get entity by entity ID.
    # @param [Fixnum] id
    # @return [Sketchup::Entity, nil]
    # @since 2.0.0
    def get_entity_by_id(id)
      model = Sketchup.active_model
      model.entities.each { |e|
        return e if e.entityID == id
      }
      model.definitions.each { |d|
        d.instances.each { |e| return e if e.entityID == id }
        d.entities.each { |e| return e if e.entityID == id }
      }
      nil
    end

    # Determine whether object is +true+ or +false+.
    # @return [Boolean]
    # @since 2.0.0
    def is_boolean?(object)
      object.is_a?(TrueClass) || object.is_a?(FalseClass)
    end

    alias is_bool? is_boolean?

  end # class << self
end # module AMS


module AMS::Lib

  # Get folder location.
  PATH = File.dirname(__FILE__).freeze
  # Get name of the main file.
  BASE = File.basename(__FILE__).freeze
  # @!visibility private
  REGISTERED_FILES = [
    'cursor.rb', 'keyboard.rb', 'main.rb', 'menu.rb', 'midi.rb',
    'multi_line_text.rb', 'ray_util.rb', 'registry.rb', 'ruby_c.rb',
    'screen.rb', 'sketchup.rb', 'sketchup_observer.rb', 'system.rb', 'timer.rb',
    'translate.rb', 'window.rb'
  ]

  class << self

    # Require all library.
    # @note All already loaded files will be ignored.
    # @return [Fixnum] number of files required.
    # @since 2.0.0
    def require_all
      count = 0
      Dir.glob( File.join(PATH, '*.{rb, rbs}') ).each { |fpath|
        base_name = File.basename(fpath)
        next if base_name == BASE || !REGISTERED_FILES.include?(base_name)
        count += 1 if Kernel.require(fpath)
      }
      count
    end

    # Load all library.
    # @note All already loaded files will be reloaded.
    # @return [Fixnum] number of files loaded.
    # @since 2.0.0
    def load_all
      count = 0
      Dir.glob( File.join(PATH, '*.{rb, rbs}') ).each { |fpath|
        base_name = File.basename(fpath)
        next if base_name == BASE || !REGISTERED_FILES.include?(base_name)
        Kernel.load fpath
        count += 1
      }
      count
    end

    # Require specific file within the library's path.
    # @note The file will not be loaded if it was already loaded.
    # @param [String] file_name
    # @return [Boolean] success
    # @since 2.0.0
    def require(file_name)
      fpath = File.join(PATH, file_name)
      Kernel.require fpath
    end

    # Load specific file within the library's path.
    # @note The file will be reloaded if it was already loaded.
    # @param [String] file_name
    # @return [Boolean] success
    # @since 2.0.0
    def load(file_name)
      fpath = File.join(PATH, file_name)
      load fpath
    end

    # Remove unregistered Ruby files from the library.
    # @return [Fixnum] Number of files deleted.
	# @since 2.1.0
    def clean_up
      count = 0
      Dir.glob( File.join(PATH, '*.{rb, rbs}') ).each { |fpath|
        base_name = File.basename(fpath)
        next if base_name == BASE || REGISTERED_FILES.include?(base_name)
        File.delete(fpath)
        count += 1
      }
      count
    end

  end # class << self
end # module AMS::Lib

unless file_loaded?(__FILE__)
  file_loaded(__FILE__)
  dir = File.dirname(__FILE__)

  ops = (RUBY_PLATFORM =~ /mswin|mingw/i) ? 'win' : 'mac'
  bit = (Sketchup.respond_to?('is_64bit?') && Sketchup.is_64bit? )? '64' : '32'
  ver = (RUBY_VERSION =~ /1.8/) ? '1.8' : '2.0'

  begin
    require File.join(dir, ops+bit, ver, 'ams_lib')
    AMS::C.find_sketchup_windows(Process.pid, Sketchup.version.to_i)
    AMS::C.enable_message_tracker(true)
  rescue LoadError
    # Avoid loading library if it isn't compatible with the current Ruby build.
  end

  begin
    require File.join(dir, ops+bit, ver, 'ams_win32_api')
  rescue LoadError
    # Avoid loading library if it isn't compatible with the current Ruby build.
  end

  begin
    require File.join(dir, ops+bit, ver, 'ams_ffi_c')
    require File.join(dir, 'ffi/ffi')
  rescue LoadError
    # Avoid loading FFI if it isn't compatible with the current Ruby build.
  end

  AMS::Lib.clean_up
  AMS::Lib.require_all
end
