module AMS::Registry

  @registry_path = nil
  @predefined_keys = {
    :hkey_class_root        => 0x80000000,
    :hkey_current_user      => 0x80000001,
    :hkey_local_machine     => 0x80000002,
    :hkey_users             => 0x80000003,
    :hkey_current_config    => 0x80000005,
  }

  class << self

    private

    def format_path(full_path, rel_to_su_reg_path, read_value)
      if rel_to_su_reg_path
        @registry_path = AMS::Sketchup.get_registry_path unless @registry_path
        full_path = File.join(@registry_path, full_path)
      end
      return if full_path.empty?
      data = full_path.split('/')
      base = @predefined_keys[data[0].downcase.to_sym]
      return unless base
      if data.size == 1
        path = ""
        name = ""
      elsif data.size == 2
        path = read_value ? "" : data[1]
        name = data[1]
      else
        path = read_value ? data[1...-1].join('/') : data[1..-1].join('/')
        name = data[-1]
      end
      # Encode to ASCII-8BIT
      path = path.gsub("/", "\\").unpack('U*').pack('C*')
      name = name.unpack('U*').pack('C*')
      return read_value ? [base, path, name] : [base, path]
    end

    public

    # Get data associated with the registry path.
    # @param [String] full_path
    # @param [Boolean] rel_to_su_reg_path Whether to acquire path relative to
    #   the registry path of the current SketchUp application or from the
    #   beginning.
    # @return [String, Fixnum, Bignum, nil] Associated data or nil if the
    #   specified path is invalid.
    # @example Read registry relative to base root:
    #   read('HKEY_CURRENT_USER/Environment/TEMP', false)
    # @example Read registry relative to SketchUp registry path:
    #   read('Application/RunCounterSU', true)
    # @since 2.0.0
    def read(full_path, rel_to_su_reg_path = true)
      data = format_path(full_path, rel_to_su_reg_path, true)
      return unless data
      res = AMS::C.read_registry(data[0], data[1], data[2])
      res = res.unpack('C*').pack('U*') if res.is_a?(String)
      res
    end

    alias get read

    # Set data associated with the registry path.
    # @param [String] full_path
    # @param [Object] value
    # @param [Boolean] rel_to_su_reg_path Whether to acquire path relative to
    #   the registry path of the current SketchUp application or from the
    #   beginning.
    # @return [Boolean] success
    # @since 2.0.0
    def write(full_path, value, rel_to_su_reg_path = true)
      data = format_path(full_path, rel_to_su_reg_path, true)
      return false unless data
      value = value.unpack('U*').pack('C*') if value.is_a?(String)
      AMS::C.write_registry(data[0], data[1], data[2], value)
    end

    alias set write

    # Delete registry key or value.
    # @param [String] full_path
    # @param [Boolean] rel_to_su_reg_path Whether to acquire path relative to
    #   the registry path of the current SketchUp application or from the
    #   beginning.
    # @return [Boolean] success
    # @since 2.0.0
    def delete(full_path, rel_to_su_reg_path = true)
      data = format_path(full_path, rel_to_su_reg_path, true)
      return false unless data
      AMS::C.delete_registry(data[0], data[1], data[2])
    end

    alias remove delete

    # Get all keys or 'folders' found in the specified registry path.
    # @param [String] full_path
    # @param [Boolean] rel_to_su_reg_path Whether to acquire path relative to
    #   the registry path of the current SketchUp application or from the
    #   beginning.
    # @return [Array<String>]
    # @since 2.0.0
    def get_keys(full_path, rel_to_su_reg_path = true)
      data = format_path(full_path, rel_to_su_reg_path, false)
      return [] unless data
      keys = AMS::C.get_registry_keys(data[0], data[1])
      for i in 0...keys.length
        keys[i] = keys[i].unpack('C*').pack('U*')
      end
      keys
    end

    # Get all values or 'files' and their data found in the specified registry
    # path.
    # @param [String] full_path
    # @param [Boolean] rel_to_su_reg_path Whether to acquire path relative to
    #   the registry path of the current SketchUp application or from the
    #   beginning.
    # @return [Hash{String => Fixnum, Bignum, String}] { value name => value_data, ... }
    # @since 2.0.0
    def get_values(full_path, rel_to_su_reg_path = true)
      data = format_path(full_path, rel_to_su_reg_path, false)
      return {} unless data
      hash = AMS::C.get_registry_values(data[0], data[1])
      encoded_hash = {}
      hash.each { |k,v|
        v = v.unpack('C*').pack('U*') if v.is_a?(String)
        encoded_hash[k.unpack('C*').pack('U*')] = v
      }
      encoded_hash
    end

  end # class << self
end # module AMS::Registry
