# This file is meant to add some ruby implementation to some C functions, which
# would have taken more effort to write something similar in C.

# @private
module AMS::C

  @observers ||= []
  @accelerators ||= {}
  @timer_procedures ||= {}
  @triggered_command = 0
  @locale = ::LanguageHandler.new('Localizable.strings')

  class << self

    private

    # Called by a c extension whenever particular event is being triggered.
    def __call_event(evt, params = [])
      process = true
      evt = evt.to_s
      update = 0
      if evt == 'swp_on_key_down' || evt == 'swp_on_key_extended' || evt == 'swp_on_key_up'
        func = RUBY_VERSION.to_f < 1.9 ? 'index' : 'key'
        vk_name = VIRTUAL_KEY_CODES.method(func).call(params[0])
        vk_char = params[2].unpack('C*').pack('U*')
        update = 1
      elsif evt == 'swo_on_caption_changed'
        caption = params[0].unpack('C*').pack('U*')
        update = 2
      end
      @observers.each { |obj|
        next unless obj.respond_to?(evt)
        if update == 1
          params[0] = vk_name.dup
          params[2] = vk_char.dup
        elsif update == 2
          params[0] = caption.dup
        end
        begin
          ret = obj.method(evt).call(*params)
          process = false if ret == 1
        rescue Exception => e
          if obj.respond_to?(evt) && evt != 'swo_error'
            __call_event(:swo_error, [e]) if (evt != 'swo_error')
          else
            puts "Exception in #{obj}.#{evt} :\n#{e}\n#{e.backtrace[0..2].join("\n")}\n"
          end
        end
      }
      process
    rescue Exception => e
      p e
    end

    # Called by a c extension to check whether the observers should be processed.
    def __get_observers_count
      @observers.size
    rescue Exception => e
      p e
    end

    # Called by a c extension whenever the menu bar is removed.
    def __find_accelerators
      @accelerators.clear
      commands = get_menu_commands(get_main_window_menu_bar)
      Sketchup.get_shortcuts.each { |command|
        shortcut, menu_path = command.split("\t")
        id = commands[menu_path]
        next if id.nil?
        keys = shortcut.split('+')
        keys[-1] << '+' if shortcut[-1,1] == '+'
        # Translate shortcut to English.
        for i in 0...keys.size
          func = RUBY_VERSION.to_f < 1.9 ? 'index' : 'key'
          str = @locale.instance_variable_get('@strings').method(func).call(keys[i])
          keys[i] = str if str
        end
        # Create accelerator.
        accel = Array.new(4)
        accel[0] = (shortcut =~ /Ctrl+/i ? 1 : 0)
        accel[1] = (shortcut =~ /Alt+/i ? 1 : 0)
        accel[2] = (shortcut =~ /Shift+/i ? 1 : 0)
        accel[3] = get_virtual_key_value_by_name(keys.last)
        @accelerators[accel] = id
      }
      nil
    rescue Exception => e
      p e
    end

    # Called by a c extension whenever there is a WM_KEYDOWN message.
    # This event is triggered only if the menu bar is removed.
    def __call_accelerator(ctrl_down, alt_down, shift_down, vk)
      accel = [ctrl_down, alt_down, shift_down, vk]
      id = @accelerators[accel]
      return false unless id
      @triggered_command = 0
      AMS::Timer.start(0, false){
        Object::Sketchup.send_action(id) if @triggered_command != id
      }
      true
    rescue Exception => e
      p e
    end

    # Called by a c extension whenever there is a WM_COMMAND message.
    # This event is triggered only if the menu bar is removed.
    def __verify_accelerator(id)
      @triggered_command = id if @triggered_command.zero?
    rescue Exception => e
      p e
    end

    # Called by timers.
    def __call_timer_proc(id, time)
      data = @timer_procedures[id]
      return false unless data
      data[2] += 1
      begin
        data[0].call(data[2], time)
      rescue Exception => e
        puts "Exception in AMS Timer ##{id}:\n#{e}\n#{e.backtrace[0..2].join("\n")}\n"
      end
      stop_timer(id) unless data[1]
      true
    rescue Exception => e
      p e
    end

    public

    def add_observer(obj)
      return false if @observers.include?(obj)
      @observers << obj
      __call_event(:swo_activate)
      true
    end

    def remove_observer(obj)
      if @observers.delete(obj)
        __call_event(:swo_deactivate)
        return true
      end
      false
    end

    def get_virtual_key_value_by_name(vk_name)
      vk_name = vk_name.to_s
      if vk_name != ' ' && vk_name != '_'
        vk_name = vk_name.to_s.downcase.gsub(/_|\s/, '')
        vk_name = ' ' if vk_name.empty?
      end
      value = VIRTUAL_KEY_CODES[vk_name]
      value = VIRTUAL_KEY_CODES_EXTENDED[vk_name] unless value
      value
    end

    def get_menu_commands(hmenu, cur_path = '')
      commands = {}
      return commands unless is_menu?(hmenu)
      commands[cur_path] = hmenu
      for i in 0...get_menu_item_count(hmenu)
        str = get_menu_item_string_by_pos(hmenu, i).to_s.unpack('C*').pack('U*')
        name = str.split("\t")[0].to_s.delete('&')
        path = cur_path.empty? ? name : cur_path + '/' + name
        id = get_menu_item_id(hmenu, i)
        if id == -1
          sub_menu = get_sub_menu(hmenu, i)
          commands.merge!(get_menu_commands(sub_menu, path))
        else
          commands[path] = id
        end
      end
      commands
    end

    def start_timer(time_out, repeat, &block)
      id = rand(10000)+1
      while @timer_procedures[id] != nil
        id = rand(10000)+1
      end
      @timer_procedures[id] = [block, repeat == true, 0]
      set_timer(id, time_out)
      id
    end

    def stop_timer(id)
      @timer_procedures.delete(id)
      kill_timer(id)
    end

  end # class << self

  VIRTUAL_KEY_CODES = {
    'lbutton'           => 0x01,
    'rbutton'           => 0x02,
    'cancel'            => 0x03,
    'mbutton'           => 0x04,
    'xbutton1'          => 0x05,
    'xbutton2'          => 0x06,
    'back'              => 0x08,
    'tab'               => 0x09,
    'clear'             => 0x0C,
    'return'            => 0x0D,
    'shift'             => 0x10,
    'control'           => 0x11,
    'menu'              => 0x12,
    'pause'             => 0x13,
    'capital'           => 0x14,
    'kana'              => 0x15,
    'handuel'           => 0x15,
    'hangul'            => 0x15,
    'junja'             => 0x17,
    'final'             => 0x18,
    'hanja'             => 0x19,
    'hanji'             => 0x19,
    'escape'            => 0x1B,
    'convert'           => 0x1C,
    'nonconvert'        => 0x1D,
    'accept'            => 0x1E,
    'modechange'        => 0x1F,
    'space'             => 0x20,
    'prior'             => 0x21,
    'next'              => 0x22,
    'end'               => 0x23,
    'home'              => 0x24,
    'left'              => 0x25,
    'up'                => 0x26,
    'right'             => 0x27,
    'down'              => 0x28,
    'select'            => 0x29,
    'print'             => 0x2A,
    'execute'           => 0x2B,
    'snapshot'          => 0x2C,
    'insert'            => 0x2D,
    'delete'            => 0x2E,
    'help'              => 0x2F,
    '0'                 => 0x30,
    '1'                 => 0x31,
    '2'                 => 0x32,
    '3'                 => 0x33,
    '4'                 => 0x34,
    '5'                 => 0x35,
    '6'                 => 0x36,
    '7'                 => 0x37,
    '8'                 => 0x38,
    '9'                 => 0x39,
    'a'                 => 0x41,
    'b'                 => 0x42,
    'c'                 => 0x43,
    'd'                 => 0x44,
    'e'                 => 0x45,
    'f'                 => 0x46,
    'g'                 => 0x47,
    'h'                 => 0x48,
    'i'                 => 0x49,
    'j'                 => 0x4A,
    'k'                 => 0x4B,
    'l'                 => 0x4C,
    'm'                 => 0x4D,
    'n'                 => 0x4E,
    'o'                 => 0x4F,
    'p'                 => 0x50,
    'q'                 => 0x51,
    'r'                 => 0x52,
    's'                 => 0x53,
    't'                 => 0x54,
    'u'                 => 0x55,
    'v'                 => 0x56,
    'w'                 => 0x57,
    'x'                 => 0x58,
    'y'                 => 0x59,
    'z'                 => 0x5A,
    'lwin'              => 0x5B,
    'rwin'              => 0x5C,
    'apps'              => 0x5D,
    'sleep'             => 0x5F,
    'numpad0'           => 0x60,
    'numpad1'           => 0x61,
    'numpad2'           => 0x62,
    'numpad3'           => 0x63,
    'numpad4'           => 0x64,
    'numpad5'           => 0x65,
    'numpad6'           => 0x66,
    'numpad7'           => 0x67,
    'numpad8'           => 0x68,
    'numpad9'           => 0x69,
    'multiply'          => 0x6A,
    'add'               => 0x6B,
    'separator'         => 0x6C,
    'subtract'          => 0x6D,
    'decimal'           => 0x6E,
    'divide'            => 0x6F,
    'f1'                => 0x70,
    'f2'                => 0x71,
    'f3'                => 0x72,
    'f4'                => 0x73,
    'f5'                => 0x74,
    'f6'                => 0x75,
    'f7'                => 0x76,
    'f8'                => 0x77,
    'f9'                => 0x78,
    'f10'               => 0x79,
    'f11'               => 0x7A,
    'f12'               => 0x7B,
    'f13'               => 0x7C,
    'f14'               => 0x7D,
    'f15'               => 0x7E,
    'f16'               => 0x7F,
    'f17'               => 0x80,
    'f18'               => 0x81,
    'f19'               => 0x82,
    'f20'               => 0x83,
    'f21'               => 0x84,
    'f22'               => 0x85,
    'f23'               => 0x86,
    'f24'               => 0x87,
    'numlock'           => 0x90,
    'scroll'            => 0x91,
    'lshift'            => 0xA0,
    'rshift'            => 0xA1,
    'lcontrol'          => 0xA2,
    'rcontrol'          => 0xA3,
    'lmenu'             => 0xA4,
    'rmenu'             => 0xA5,
    'browserback'       => 0xA6,
    'browserforward'    => 0xA7,
    'browserrefresh'    => 0xA8,
    'browserstop'       => 0xA9,
    'browsersearch'     => 0xAA,
    'browserfavorites'  => 0xAB,
    'browserhome'       => 0xAC,
    'volumemute'        => 0xAD,
    'volumedown'        => 0xAE,
    'volumeup'          => 0xAF,
    'medianexttrack'    => 0xB0,
    'mediaprevtrack'    => 0xB1,
    'mediastop'         => 0xB2,
    'mediaplaypause'    => 0xB3,
    'launchmail'        => 0xB4,
    'launchmediaselect' => 0xB5,
    'launchapp1'        => 0xB6,
    'launchapp2'        => 0xB7,
    'oem1'              => 0xBA, # US keyboard : ;
    'oemplus'           => 0xBB, # US keyboard + =
    'oemcomma'          => 0xBC, # US keyboard < ,
    'oemminus'          => 0xBD, # US keyboard _ -
    'oemperiod'         => 0xBE, # US keyboard > .
    'oem2'              => 0xBF, # US keyboard ? /
    'oem3'              => 0xC0, # US keyboard ~ `
    'oem4'              => 0xDB, # US keyboard { [
    'oem5'              => 0xDC, # US keyboard | \
    'oem6'              => 0xDD, # US keyboard } ]
    'oem7'              => 0xDE, # US keyboard " '
    'oem8'              => 0xDF,
    'attn'              => 0xF6,
    'crsel'             => 0xF7,
    'exsel'             => 0xF8,
    'ereof'             => 0xF9,
    'play'              => 0xFA,
    'zoom'              => 0xFB,
    'pa1'               => 0xFD,
    'oemclear'          => 0xFE
  }

  VIRTUAL_KEY_CODES_EXTENDED = {
    'backspace'         => 0x08,
    'enter'             => 0x0D,
    'ctrl'              => 0x11,
    'alt'               => 0x12,
    'break'             => 0x13,
    'capslock'          => 0x14,
    'esc'               => 0x1B,
    'spacebar'          => 0x20,
    ' '                 => 0x20,
    'pageup'            => 0x21,
    'pagedown'          => 0x22,
    '←'                 => 0x25,
    '↑'                 => 0x26,
    '→'                 => 0x27,
    '↓'                 => 0x28,
    'printscreen'       => 0x2C,
    'prtscn'            => 0x2C,
    'sysrq'             => 0x2C,
    ')'                 => 0x30,
    'closeparenthese'   => 0x30,
    '!'                 => 0x31,
    'exclamation'       => 0x31,
    'exclamationmark'   => 0x31,
    '@'                 => 0x32,
    'at'                => 0x32,
    'atsign'            => 0x32,
    '#'                 => 0x33,
    'number'            => 0x33,
    'numbersign'        => 0x33,
    'hash'              => 0x33,
    '$'                 => 0x34,
    'dollar'            => 0x34,
    'dollarsign'        => 0x34,
    '%'                 => 0x35,
    'percent'           => 0x35,
    '^'                 => 0x36,
    'caret'             => 0x36,
    '&'                 => 0x37,
    'ampersand'         => 0x37,
    '*'                 => 0x38,
    'asterisk'          => 0x38,
    '('                 => 0x39,
    'openparenthese'    => 0x39,
    'numpad*'           => 0x6A,
    'numpad+'           => 0x6B,
    'numpad-'           => 0x6D,
    'numpad.'           => 0x6E,
    'numpad/'           => 0x6F,
    'scrolllock'        => 0x91,
    'scrlk'             => 0x91,
    'lctrl'             => 0xA2,
    'rctrl'             => 0xA3,
    'lalt'              => 0xA4,
    'ralt'              => 0xA5,
    'mute'              => 0xAD,
    'medianext'         => 0xB0,
    'mediaprev'         => 0xB1,
    'mediatoggleplay'   => 0xB3,
    'mail'              => 0xB4,
    'mediaselect'       => 0xB5,
    'app1'              => 0xB6,
    'app2'              => 0xB7,
    'colon'             => 0xBA,
    'semicolon'         => 0xBA,
    ':'                 => 0xBA,
    ';'                 => 0xBA,
    'plus'              => 0xBB,
    'equals'            => 0xBB,
    '+'                 => 0xBB,
    '='                 => 0xBB,
    'comma'             => 0xBC,
    ','                 => 0xBC,
    '<'                 => 0xBC,
    'minus'             => 0xBD,
    'dash'              => 0xBD,
    'underscore'        => 0xBD,
    '_'                 => 0xBD,
    '-'                 => 0xBD,
    'period'            => 0xBE,
    '.'                 => 0xBE,
    '>'                 => 0xBE,
    'question'          => 0xBF,
    'questionmark'      => 0xBF,
    'slash'             => 0xBF,
    '?'                 => 0xBF,
    '/'                 => 0xBF,
    'tilde'             => 0xC0,
    '~'                 => 0xC0,
    '`'                 => 0xC0,
    'openbrace'         => 0xDB,
    'openbracket'       => 0xDB,
    '{'                 => 0xDB,
    '['                 => 0xDB,
    'verticalbar'       => 0xDC,
    'backslash'         => 0xDC,
    '|'                 => 0xDC,
    "\\"                => 0xDC,
    'closebrace'        => 0xDD,
    'closebracket'      => 0xDD,
    '}'                 => 0xDD,
    ']'                 => 0xDD,
    'quote'             => 0xDE,
    'quotaion'          => 0xDE,
    'quotaionmark'      => 0xDE,
    'apostrophe'        => 0xDE,
    "\'"                => 0xDE,
    "\""                => 0xDE
  }

end # module AMS::C
