module AMS::Sketchup

  @is_64bit = ( ::Sketchup.respond_to?('is_64bit?') && ::Sketchup.is_64bit? )
  @dialogs ||= []
  @toolbars ||= []
  @dialogs_to_ignore ||= []
  @su_controlled_dialogs = [
    'Model Info', 'Entity Info', 'Materials', 'Components', 'Styles', 'Layers',
    'Outliner', 'Scenes', 'Shadow Settings', 'Fog', 'Match Photo',
    'Soften Edges', 'Instructor'
  ]

  class << self

    # Determine whether current SketchUp application is 32bit.
    # @return [Boolean]
    # @since 2.0.0
    def is_32bit?
      !@is_64bit
    end

    # Determine whether current SketchUp application is 64bit.
    # @return [Boolean]
    # @since 2.0.0
    def is_64bit?
      @is_64bit
    end

    # Get module handle to the current SketchUp application.
    # @return [Fixnum]
    # @since 2.0.0
    def get_module_handle
      AMS::C.get_module_handle
    end

    # Get full path to the current SketchUp executable.
    # @return [String]
    # @since 2.0.0
    def get_executable_path
      path = AMS::C.get_executable_path().unpack('C*').pack('U*')
      path.gsub!(/\\/, '/')
      path
    end

    # Get registry path of the current SketchUp application.
    # @return [String]
    # @since 2.0.0
    def get_registry_path
      version = Sketchup.version.to_i
      case version
      when 5
       "HKEY_CURRENT_USER/Software/Google/Google SketchUp5/"
      when 6..8
        "HKEY_CURRENT_USER/Software/Google/SketchUp#{version}/"
      else
        "HKEY_CURRENT_USER/Software/SketchUp/SketchUp 20#{version}/"
      end
    end

    # Get process id of the current SketchUp application.
    # @return [Fixnum]
    # @since 2.0.0
    def get_process_id
      AMS::C.get_process_id
    end

    # Get thread id of the current SketchUp application.
    # @return [Fixnum]
    # @since 2.0.0
    def get_thread_id
      AMS::C.get_thread_id
    end

    # Get handle to the main window of the current SketchUp application.
    # @return [Fixnum]
    # @since 2.0.0
    def get_main_window
      AMS::C.get_main_window
    end

    # Get handle to the menu bar of the main window.
    # @return [Fixnum]
    # @since 2.0.0
    def get_menu_bar
      AMS::C.get_main_window_menu_bar
    end

    # Get handle to the view window of the main window.
    # @return [Fixnum]
    # @since 2.0.0
    def get_viewport
      AMS::C.get_viewport
    end

    # Get handle to the status bar of the main window.
    # @return [Fixnum]
    # @since 2.0.0
    def get_status_bar
      AMS::C.get_status_bar
    end

    # Get handle to the scenes bar of the main window.
    # @return [Fixnum]
    # @since 2.0.0
    def get_scenes_bar
      AMS::C.get_scenes_bar
    end

    # Get handle to the four toolbar containers of the main window.
    # @return [Array<Fixnum>] +[top toolbar, bottom toolbar, left toolbar, right toolbar]+
    # @since 2.0.0
    def get_toolbar_containers
      AMS::C.get_control_bars
    end

    # Set main window full screen.
    # @param [Boolean] state
    # @param [Fixnum] reset_mode Whether to set restored (0), set maximized (1),
    #   or set to original placement (2) when main window is unset full screen.
    # @return [Boolean] success
    # @since 2.0.0
    def switch_full_screen(state, reset_mode = 2)
      AMS::C.set_main_window_full_screen(state, reset_mode.to_i)
    end

    # Determine whether main window is full screen.
    # @return [Boolean]
    # @since 2.0.0
    def is_full_screen?
      AMS::C.is_main_window_full_screen?
    end

    # Set main window maximized.
    # @return [Boolean] success
    # @since 2.0.0
    def maximize
      return false if is_maximized?
      AMS::C.set_main_window_full_screen(false, 1)
      AMS::C.show_window(AMS::C.get_main_window(), 3)
      true
    end

    # Determine whether main window is maximized.
    # @return [Boolean]
    # @since 2.0.0
    def is_maximized?
      return false if AMS::C.is_main_window_full_screen?
      AMS::C.is_window_maximized?(AMS::C.get_main_window())
    end

    # Set main window minimized.
    # @return [Boolean] success
    # @since 2.0.0
    def minimize
      return false if is_minimized?
      AMS::C.show_window(AMS::C.get_main_window(), 3)
      true
    end

    # Determine whether main window is minimized.
    # @return [Boolean]
    # @since 2.0.0
    def is_minimized?
      AMS::C.is_window_minimized?(AMS::C.get_main_window())
    end

    # Set main window to normal placement.
    # @return [Boolean] success
    # @since 2.0.0
    def restore
      return false if is_restored?
      AMS::C.set_main_window_full_screen(false, 0)
      AMS::C.show_window(AMS::C.get_main_window(), 9)
      true
    end

    # Determine whether main window is restored.
    # @return [Boolean]
    # @since 2.0.0
    def is_restored?
      return false if AMS::C.is_main_window_full_screen?
      return false if AMS::C.is_window_minimized?(AMS::C.get_main_window())
      return false if AMS::C.is_window_maximized?(AMS::C.get_main_window())
      true
    end

    # Set main window active.
    # @return [Boolean] success
    # @since 2.0.0
    def activate
      AMS::C.bring_window_to_top(AMS::C.get_main_window())
    end

    alias bring_to_top activate

    # Set/Remove menu bar.
    # @param [Boolean] state
    # @return [Boolean] success
    # @since 2.0.0
    def set_menu_bar(state)
      AMS::C.set_main_window_menu_bar(state)
    end

    # Determine whether menu bar is set.
    # @return [Boolean]
    # @since 2.0.0
    def is_menu_bar_set?
      AMS::C.is_main_window_menu_bar_set?
    end

    # Set/Remove viewport border, a thin edge surrounding the view.
    # @param [Boolean] state
    # @return [Boolean] success
    # @since 2.0.0
    def set_viewport_border(state)
      AMS::C.set_viewport_border(state)
    end

    # Determine whether viewport border is set.
    # @return [Boolean]
    # @since 2.0.0
    def is_viewport_border_set?
      AMS::C.is_viewport_border_set?
    end

    # Show/Hide status bar.
    # @param [Boolean] state
    # @param [Boolean] refresh Whether to update after changing state.
    # @return [Boolean] success
    # @since 2.0.0
    def show_status_bar(state, refresh = true)
      AMS::C.set_status_bar_visible(state, refresh)
    end

    # Determine whether status bar is visible.
    # @return [Boolean]
    # @since 2.0.0
    def is_status_bar_visible?
      AMS::C.is_status_bar_visible?
    end

    # Show/Hide scenes bar.
    # @param [Boolean] state
    # @param [Boolean] refresh Whether to update after changing state.
    # @return [Boolean] success
    # @since 2.0.0
    def show_scenes_bar(state, refresh = true)
      return false if state && AMS::C.is_scenes_bar_empty?()
      AMS::C.set_scenes_bar_visible(state, refresh)
    end

    # Determine whether scenes bar is visible.
    # @return [Boolean]
    # @since 2.0.0
    def is_scenes_bar_visible?
      AMS::C.is_scenes_bar_visible?
    end

    # Determine whether scenes bar is filled. Scenes bar is filled if there
    # is at least one page in the model.
    # @return [Boolean]
    # @since 2.0.0
    def is_scenes_bar_filled?
      AMS::C.is_scenes_bar_filled?()
    end

    # Determine whether scenes bar is empty. Scenes bar is empty if there are
    # no pages in the model.
    # @return [Boolean]
    # @since 2.0.0
    def is_scenes_bar_empty?
      AMS::C.is_scenes_bar_empty?()
    end

    # Show/Hide toolbar container(s).
    # @param [Fixnum] bar 1 - top, 2 - bottom, 3 - left, 4 - right,
    #   5 - all filled, 6 - all.
    # @param [Boolean] state
    # @param [Boolean] refresh Whether to update after changing state.
    # @return [Boolean] success
    # @since 2.0.0
    def show_toolbar_container(bar, state, refresh = true)
      bar = AMS.clamp(bar.to_i, 1, 6)
      case bar
      when 1..4
        AMS::C.set_toolbar_container_visible(bar, state, refresh)
      when 5
        change = false
        for i in 1..4
          if AMS::C.is_toolbar_container_filled?(i)
            change = true if AMS::C.set_toolbar_container_visible(i, state, refresh)
          end
        end
        change
      else
        change = false
        for i in 1..4
          change = true if AMS::C.set_toolbar_container_visible(i, state, refresh)
        end
        change
      end
    end

    # Determine whether toolbar container(s) is/are visible.
    # @param [Fixnum] bar 1 - top, 2 - bottom, 3 - left, 4 - right,
    #   5 - at least one, 6 - all, 7 - all filled.
    # @return [Boolean]
    # @since 2.0.0
    def is_toolbar_container_visible?(bar)
      bar = AMS.clamp(bar.to_i, 1, 7)
      case bar
      when 1..4
        AMS::C.is_toolbar_container_visible?(bar)
      when 5
        for i in 1..4
          return true if AMS::C.is_toolbar_container_visible?(i)
        end
        false
      when 6
        for i in 1..4
          return false unless AMS::C.is_toolbar_container_visible?(i)
        end
        true
      else
        count = 0
        for i in 1..4
          return false if AMS::C.is_toolbar_container_filled?(i) && !AMS::C.is_toolbar_container_visible?(i)
        end
        true
      end
    end

    # Determine whether toolbar container(s) is/are filled.
    # @param [Fixnum] bar 1 - top, 2 - bottom, 3 - left, 4 - right,
    #   5 - at least one, 6 - all.
    # @return [Boolean]
    # @since 2.0.0
    def is_toolbar_container_filled?(bar)
      bar = AMS.clamp(bar.to_i, 1, 6)
      case bar
      when 1..4
        AMS::C.is_toolbar_container_filled?(bar)
      when 5
        for i in 1..4
          return true if AMS::C.is_toolbar_container_filled?(i)
        end
        false
      else
        for i in 1..4
          return false unless AMS::C.is_toolbar_container_filled?(i)
        end
        true
      end
    end

    # Determine whether toolbar container(s) is/are empty.
    # @param [Fixnum] bar 1 - top, 2 - bottom, 3 - left, 4 - right,
    #   5 - at least one, 6 - all.
    # @return [Boolean]
    # @since 2.0.0
    def is_toolbar_container_empty?(bar)
      bar = AMS.clamp(bar.to_i, 1, 6)
      case bar
      when 1..4
        AMS::C.is_toolbar_container_empty?(bar)
      when 5
        for i in 1..4
          return true if AMS::C.is_toolbar_container_empty?(i)
        end
        false
      else
        for i in 1..4
          return false unless AMS::C.is_toolbar_container_empty?(i)
        end
        true
      end
    end

    # Get upper-left and lower-right corners of the view window in screen
    # coordinates, relatives to the upper-left corner of the screen.
    # @return [Array<Fixnum>] +[x1,y1, x2,y2]+
    # @since 2.0.0
    def get_viewport_rect
      AMS::C.get_viewport_rect()
    end

    # Get upper-left corner of the view window in screen coordinates.
    # @return [Array<Fixnum>] +[x,y]+
    # @since 2.0.0
    def get_viewport_origin
      AMS::C.get_viewport_rect()[0,2]
    end

    # Get width and height of the view window.
    # @note This is same as <tt>view.vp_width</tt> and <tt>view.vp_height</tt>.
    # @return [Array<Fixnum>] +[width, height]+
    # @since 2.0.0
    def get_viewport_size
      r = AMS::C.get_viewport_rect()
      [ r[2] - r[0], r[3] - r[1] ]
    end

    # Get viewport center in screen coordinates.
    # @return [Array<Fixnum>] +[x,y]+
    # @since 2.0.0
    def get_viewport_center
      r = AMS::C.get_viewport_rect()
      [((r[2] - r[0])*0.5).round, ((r[3] - r[1])*0.5).round]
    end

    # Show/Hide all dialogs.
    # @note Ignored windows won't be included in this operation.
    # @param [Boolean] state
    # @return [Fixnum] number of dialogs shown or hidden.
    # @since 2.0.0
    def show_dialogs(state)
      count = 0
      su_dialogs = []
      if state
        @dialogs.each { |handle|
          next unless AMS::C.is_window?(handle)
          next if @dialogs_to_ignore.include?(handle) || AMS::C.is_window_visible?(handle)
          caption = AMS::Window.get_text(handle)
          if @su_controlled_dialogs.include?(caption)
            su_dialogs << handle
          else
            AMS::C.show_window(handle, 8) # SW_SHOWNA
          end
          count += 1
        }
        @dialogs.clear
        unless su_dialogs.empty?
          Sketchup.send_action(21926)
          UI.start_timer(0.1, false){
            su_dialogs.each { |handle|
              next unless AMS::C.is_window?(handle)
              next if AMS::C.is_window_visible?(handle)
              Sketchup.send_action(21926)
              break
            }
          }
        end
      else
        visible_dialogs = AMS::C.get_visible_dialogs()
        @dialogs.concat visible_dialogs
        visible_dialogs.each { |handle|
          next if @dialogs_to_ignore.include?(handle)
          caption = AMS::Window.get_text(handle)
          if @su_controlled_dialogs.include?(caption)
            su_dialogs << handle
          else
            AMS::C.show_window(handle, 0) # SW_HIDE
          end
          count += 1
        }
        unless su_dialogs.empty?
          Sketchup.send_action(21926)
          UI.start_timer(0.1, false){
            su_dialogs.each { |handle|
              next unless AMS::C.is_window?(handle)
              next unless AMS::C.is_window_visible?(handle)
              Sketchup.send_action(21926)
              break
            }
          }
        end
      end
      count
    end

    # Close all dialogs.
    # @note Ignored windows won't be included in this operation.
    # @return [Fixnum] number of dialogs closed.
    # @since 2.0.0
    def close_dialogs
      count = 0
      (@dialogs + AMS::C.get_visible_dialogs()).each { |handle|
        next unless AMS::C.is_window?(handle)
        next if @dialogs_to_ignore.include?(handle)
        AMS::C.close_window(handle)
        count += 1
      }
      @dialogs.clear
      count
    end

    # Include dialog in the {show_dialogs} and {close_dialogs} operations.
    # @note By default, all dialogs are included in the show/hide/close dialogs
    #   operation. This method is used to remove dialog from the ignore list.
    # @param [Fixnum] handle A handle to the window.
    # @return [Boolean] success
    # @since 2.0.0
    def include_dialog(handle)
      @dialogs_to_ignore.delete(handle) ? true : false
    end

    # Elude dialog from the {show_dialogs} and {close_dialogs} operations.
    # @param [Fixnum] handle A handle to the window.
    # @return [Boolean] success
    # @since 2.0.0
    def ignore_dialog(handle)
      # Verify given handle is not already in the ignore list.
      return false if @dialogs_to_ignore.include?(handle)
      # Verify given handle is a window.
      return false unless AMS::C.is_window?(handle)
      # Verify given handle is part of the current SketchUp process.
      return false if AMS::C.get_window_process_id(handle) != AMS::C.get_process_id()
      # Verify given handle is a pop-up window.
      return false if AMS::C.get_window_class_name(handle) != '#32770'
      # Add handle to the ignore list if it passed all the tests.
      @dialogs_to_ignore << handle
      true
    end

    # Show/Hide all floating toolbars.
    # @param [Boolean] state
    # @return [Fixnum] number of floating toolbars shown or hidden.
    # @since 2.0.0
    def show_toolbars(state)
      count = 0
      if state
        @toolbars.each { |handle|
          next unless AMS::C.is_window?(handle)
          AMS::C.show_window(handle, 8) # SW_SHOWNA
          count += 1
        }
        @toolbars.clear
      else
        visible_toolbars = AMS::C.get_visible_toolbars()
        @toolbars.concat visible_toolbars
        visible_toolbars.each { |handle|
          AMS::C.show_window(handle, 0) # SW_HIDE
          count += 1
        }
      end
      count
    end

    # Close all floating toolbars.
    # @return [Fixnum] number of floating toolbars closed.
    # @since 2.0.0
    def close_toolbars
      count = 0
      (@toolbars + AMS::C.get_visible_toolbars()).each { |handle|
        next unless AMS::C.is_window?(handle)
        AMS::C.close_window(handle)
        count += 1
      }
      @toolbars.clear
      count
    end

    # Determine whether SketchUp main window is active.
    # @return [Boolean]
    # @since 2.0.0
    def is_main_window_active?
      AMS::C.get_active_window() == AMS::C.get_main_window()
    end

    # Determine whether the current active window belongs to the current
    # SketchUp application.
    # @return [Boolean]
    # @since 2.0.0
    def is_active?
      handle = AMS::C.get_active_window()
      AMS::C.get_window_process_id(handle) == AMS::C.get_process_id()
    end

    # Find a handle to the window with the specified caption.
    # @note This function will irritate through the windows of the current
    #   SketchUp process only. All windows belong to a different process or a
    #   different SketchUp application will not be searched.
    # @return [Fixnum, nil] A handle to the window if successful.
    # @since 2.0.0
    def find_window_by_caption(caption)
      AMS::C.find_window_by_caption(caption)
    end

    # Get all pop-up windows of the current SketchUp application.
    # @note Ignored dialogs are included in the list.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_dialogs
      AMS::C.get_dialogs() - @dialogs_to_ignore
    end

    # Get all used pop-up windows of the current SketchUp application.
    # @note Ignored dialogs are included in the list.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_active_dialogs
      (AMS::C.get_visible_dialogs() + @dialogs - @dialogs_to_ignore).uniq
    end

    # Get all visible pop-up windows of the current SketchUp application.
    # @note Ignored dialogs are included in the list.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_visible_dialogs
      AMS::C.get_visible_dialogs() - @dialogs_to_ignore
    end

    # Get all floating toolbars of the current SketchUp application.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_toolbars
      AMS::C.get_toolbars()
    end

    # Get all used floating toolbars.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_active_toolbars
      (AMS::C.get_visible_toolbars() + @toolbars).uniq
    end

    # Get all visible floating toolbars of the current SketchUp application.
    # @return [Array<Fixnum>] An array of window handles.
    # @since 2.0.0
    def get_visible_toolbars
      AMS::C.get_visible_toolbars()
    end

    # Get main window title text.
    # @return [String]
    # @since 2.0.0
    def get_caption
      AMS::C.get_window_text(AMS::C.get_main_window()).unpack('C*').pack('U*')
    end

    # Set main window title text.
    # @param [String] caption
    # @return [Boolean] success
    # @since 2.0.0
    def set_caption(caption)
      AMS::C.set_window_text(AMS::C.get_main_window(), caption.to_s.unpack('U*').pack('C*'))
    end

    # Refresh current SketchUp application.
    # @return [void]
    # @since 2.0.0
    def refresh
      AMS::C.refresh_window(AMS::C.get_main_window())
    end

    # Close current SketchUp application.
    # @note This behaves the same as clicking the 'X' button.
    # @return [void]
    # @since 2.0.0
    def close
      AMS::C.close_window(AMS::C.get_main_window())
    end

    # Add object to the observers list.
    # @note An observer can be a class, module, or a class instance. Your
    #   observer will work as long as the callback methods are public.
    # @note Your observer is not supposed to contain every callback method from
    #   the observers list. You may include/elude those as you wish.
    # @note A unique extension +swp+ or +swo+ is added in front of each observer
    #   method. SWP stands for Sketchup Window Procedure, and SWO stands for
    #   Sketchup Window Observer. Both SWP and SWO events are capable to monitor
    #   window messages, but SWP events are also capable to make decisions to
    #   the message, whether or not the message should interact with SketchUp
    #   window. If the return value for the SWP callback method is 1, then the
    #   message associated with the event will not interact with SketchUp
    #   window. For example, returning 1 in the swp_on_key_down event will
    #   prevent the key from interacting with SketchUp window, which means any
    #   shortcut associated with such key will not be triggered. SWP events may
    #   come handy for extensions that want more control over SketchUp.
    # @param [Object] object
    # @return [Boolean] success
    # @example Adding observer from class instance:
    #   class MySketchupObserver
    #     def swo_on_switch_full_screen(state)
    #       puts 'Main window set full screen!' if state
    #     end
    #   end # module MySketchupObserver
    #   AMS::Sketchup.add_observer(MySketchupObserver.new)
    # @example Adding observer from module:
    #   module MySketchupObserver
    #     def self.swo_on_switch_full_screen(state)
    #       puts 'Main window set full screen!' if state
    #     end
    #   end # module MySketchupObserver
    #   AMS::Sketchup.add_observer(MySketchupObserver)
    # @example Another way to add observer from module:
    #   module MySketchupObserver
    #     class << self
    #       def swo_on_switch_full_screen(state)
    #         puts 'Main window set full screen!' if state
    #       end
    #     end # class << self
    #   end # module MySketchupObserver
    #   AMS::Sketchup.add_observer(MySketchupObserver)
    # @example You may also add observer by inheriting, but it's not required:
    #   class MySketchupObserver < AMS::SketchupObserver
    #     def swo_on_switch_full_screen(state)
    #       puts 'Main window set full screen!' if state
    #     end
    #   end # module MySketchupObserver
    #   AMS::Sketchup.add_observer(MySketchupObserver.new)
    # @see AMS::SketchupObserver
    # @since 2.0.0
    def add_observer(object)
      AMS::C.add_observer(object)
    end

    # Remove object from the observers list.
    # @param [Object] object
    # @return [Boolean] success
    # @since 2.0.0
    def remove_observer(object)
      AMS::C.remove_observer(object)
    end

  end # class << self
end # module AMS::Sketchup
